(defclass pov ()
  ((p1 :initarg :instr :accessor instr)
   (p2 :initarg :start :accessor start)
   (p3 :initarg :end :accessor end)))
 
(defmethod print-object ((obj pov) stream)
  (let ((*print-case* ':downcase))
    (format stream "#< ~A, start: ~A, end: ~A >~%" (instr obj)
            (start obj)
            (end obj))))

(defmethod params ((obj pov))
  (list (instr obj)(start obj)(end obj)))
  
(defclass spline ()
  ((spline-type :initarg :type :initform 'linear_spline :accessor type)
   (spline-env :initarg :env :initform '(0 (0 0) 1 (1 1)) :accessor env)))

(defmethod print-object ((obj spline) stream)
  (let ((*print-case* ':downcase))
    (format stream "#< spline, type: ~A, env: ~A >~%"
            (type obj)
            (env obj))))

(defun pov-format-list (plist)
  (let* ((out (format nil "~(~A~)( " (string (first plist))))
         (str "")
         )
    (loop for i in (cdr plist) 
          for j from 1 below (length plist) do
          (setf str (cond ((< j (1- (length plist))) "~A, ")
                          (t "~A~%")))
          (setf out (concatenate 'string 
                                 out 
                                 (format nil str
                                         (cond ((listp i) (format nil "~%~5T~A" (pov-format-vector i)))
                                               ((equalp 'spline (type-of i)) 
                                                (format nil "~%~5T~A" (pov-format-spline i)))
                                               ((symbolp i) (format nil "~(~A~)" i))
                                               ((integerp i) i)
                                               (t (format nil "~,4F" i))
                                               )))))
    (concatenate 'string out (format nil "~5T)~%"))))

(defun pov-format-vector (plist)
  (let* ((out "< ")
         (str "")
         )
    (loop for i in plist 
          for j from 0 below (length plist) do
          (setf str (cond ((< j (1- (length plist))) "~A, ")
                          (t "~A")))
          (setf out (concatenate 'string 
                                 out 
                                 (format nil str
                                         (cond ((listp i) (format nil "~A" (pov-format-vector i)))
                                               ((integerp i) i)
                                               (t (format nil "~,4F" i))
                                                ))
                                 )))
    (concatenate 'string out " >")))

(defun pov-format-spline (spline)
  (let* ((out (format nil "spline { ~(~A~)~%~10T" (string (type spline))))
         (str "}")
         (env (env spline))
         (env-str ""))
    (loop for i in env 
          for j from 0 do
          (setf env-str (concatenate 'string env-str
                                     (format nil (if (oddp j) "~A~%~10T" "~A, ")
                                             (cond  
                                              ((listp i) (format nil "~A" (pov-format-vector i))) 
                                              ((integerp i) i)
                                              (t (format nil "~,4F" i))
                                              )))))
    (setf out (concatenate 'string out env-str str))
    out))

(defmacro defpovray (name slots)
  `(progn
     (defclass ,name (pov)
       ,(mapcar #'(lambda (i) (if (listp i) 
                                (list (first i) ':initarg  
                                      (intern (string (first i)) :keyword)
                                      ':initform (second i))
                                (list i ':initarg 
                                      (intern (string i) :keyword))))
                slots))
     (defmethod print-object ((obj ,name) stream)
       (format stream "#< ~A, start: ~A, end: ~A >~%" (instr obj)
            (start obj)
            (end obj)))
     (defmethod params ((obj ,name))
       ,(append '(list (instr obj)(start obj)(end obj))
                (mapcar #'(lambda (i) `(slot-value obj ',(if (listp i) (first i) i)))
                        slots)))
     (find-class ',name)))

#|
(defmacro new (name &rest args)
  `(make-instance ',name ,@args))
|#

(defun pov-events (outfile &rest args)
  (let* ((out (sort (loop for i in args append (if (listp i) i (list i)))
	        #'(lambda (a b) (< (start a)(start b))))))
    (with-open-file (fil outfile :direction :output :if-exists :supersede)
      (loop for i in out do
            (format fil "~A~%" (pov-format-list (params i))))
      )
    (format nil "~A" outfile)))


